/* Pace Serial Message Decoder, version 0.05

   by Rick Murray, 2006/07/03


   URI  : http://www.heyrick.co.uk/ricksworld/digibox/
   Email: heyrick -at- merseymail -dot- com


   Assembles with Borland TurboC++ v1.0 (a FREE download from Borland!).


   Requires DIRECT access to IRQ controller and serial port. Thus, it
   probably won't work on NT or XP...
   It makes a 16-bit DOS program so it should work on DOS 3.3 or later
   with UART hardware and an 80286 or later. Get an old 486 laptop or
   some sort of clunky AT-era clone? That'll do fine! You don't even
   need much memory! :-)

   
   You may modify this program as you see fit, and I encourage you to
   send me your updates. HOWEVER PLEASE NOTE THAT YOU DO NOT HAVE THE
   RIGHT TO (RE)DISTRIBUTE MODIFIED VERSIONS without first obtaining
   my permission. Just ask...
*/



/* Inclusions */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <time.h>
#include "dos.h"
#include "conio.h"



/* C still hasn't fixed on what TRUE and FALSE are. Some say TRUE is a
   non-zero valid, some say TRUE is -1 (spot a BASIC coder!). C doesn't
   have an opinion on this; so we have to define them ourselves! */
#define TRUE  1
#define FALSE 0



/* Global variables */
char appvers[] = "0.05";
char appdate[] = "2006/07/03";
long port = 0;         /* Is filled with address to serial port */
char buffer[1024];     /* Ring buffer memory block */
int  bufferhead = 0;   /* Ring buffer head (incoming writes here) */
int  buffertail = 0;   /* Ring buffer tail (we read from here) */
char command[512];     /* Command string for broken-down data chunks */
FILE *fp = NULL;       /* Logfile pointer */
struct text_info ti;   /* For scrolling "other messages" box */



/* Function prototypes */
void ctrlc_catch(int *reglist);
void restore_vector(void);
void serial_senddata(char *data);
void serial_writebyte(int mybyte);
int  buffer_ready(void);
int  buffer_readstring(void);
void decode_command(void);
void cursor_off(void);
void cursor_on(void);
void redraw_screen(void);
void write_raw_data(void);
void write_channel_info(char *info, int type);
void write_programme_info(char *info, int type);
void write_programme_description(char *info);
void write_misc_status(char *info);
void display_help(void);
void get_true_version(void);
void dump_to_printer(void);

void interrupt (*oldserial)(); /* old interrupt function pointer */
void interrupt serial_event(); /* interrupt prototype */



/* Herewith the code. I *HATED* PASCAL's liking of the 'main' function at
   the end, so you'll instead find it here. Right at the start. :-) */
int main(void)
{
   int  key = 0;
   int  portcount = 0;
   int  port1addr = 0;
   int  port2addr = 0;
   char activity[] = "|/-\\";
   int  activone = 0;
   char plotwhat[2] = { 'x' , ((LIGHTGRAY << 4) + LIGHTCYAN) } ;
   time_t timer;
   clock_t tick;
   struct tm *tblock;

   /* Print introduction */
   textcolor(WHITE);     cprintf("\r\n\r\nPace Serial Message Decoder");
   textcolor(LIGHTCYAN); cprintf(" - %s     %s\r\n", appvers, appdate);
   textcolor(LIGHTGREEN);cputs("by Rick Murray  ");
   textcolor(LIGHTBLUE); cprintf("http://www.heyrick.co.uk/digibox/\r\n\r\n");
   /* the shorter URI points to a self-updating link to the main
      Digibox content, it is only there so that I don't have to
      keep typing the /ricksworld bit - and so you don't have
      to keep remembering it! :-)
   */

   /* Search for serial ports */
   portcount = peek(0x0040, 0x0010); /* Equipment list pointer */
   portcount = ((portcount >> 9) && 3);

   if (portcount == 0)
   {
     /* Waaah! */
      textcolor(LIGHTMAGENTA);
      cprintf("This computer does not appear to have any serial ports!\r\n");
      textcolor(LIGHTGRAY);
      cprintf("\r\n");
      exit(0);
   }

   /* Interrogate DOS equip-seg for COM ports */
   port1addr = peek(0x0040, 0x000);
   if (portcount > 1)
      port2addr = peek(0x0040, 0x002);

   if (portcount > 1)
   {
      /* Get the desired port from the user */
      port = 0;
      do
      {
         textcolor(LIGHTCYAN);
         cputs("Please choose: COM[");
         textcolor(WHITE);
         putch('1');
         textcolor(LIGHTCYAN);
         cputs("] or COM[");
         textcolor(WHITE);
         putch('2');
         textcolor(LIGHTCYAN);
         cputs("]? ");

         /* Loop until a key pressed */
         while ( !kbhit() );
         key = getch();

         /* Choices */
         if (key == '1') port = port1addr; /* usu. 0x03F8 */
         if (key == '2') port = port2addr; /* usu. 0x02F8 */

         /* Default */
         if ((key == 13) || (key == 32))
         {
            port = port1addr; /* default to COM1 */
	    key = '1';
         }

         /* Escape */
         if (key == 27)
         {
	    textcolor(LIGHTMAGENTA);
	    cprintf("Aborted, bye bye!\r\n\r\n");
	    textcolor(LIGHTGRAY);
	    exit(0);
         }

         /* Something invalid */
         if (port == 0)
         {
	    textcolor(LIGHTMAGENTA);
	    cprintf("Invalid input, '1' or '2' please!\r\n");
         }
      } while (port == 0);
   }
   else
   {
      /* Only the one... */
      textcolor(LIGHTCYAN);
      cputs("There appears to be only one serial port, so using ");
      port = port1addr;
      key = '1';
   }

   /* Visual feedback */
   cprintf("COM%c:\r\n", key);
   textcolor(LIGHTGREEN);
   cprintf("Linking into serial port (addr=0x%04X)", port);


   /* Swap interrupt vectors */
   oldserial  = getvect(0x0C);
   atexit(restore_vector);
   setvect(0x0C, serial_event);


   /* Set serial port status - direct hardware pokage! <g> */
   outp((port + 3), 131); /* 8N1, DLAB active */
   outp((port + 0), 2);   /* for 57600 */
   outp((port + 1), 0);   /* high byte of baud divisor, not used */
   outp((port + 3), 3);   /* 8N1, DLAB inactive */
   outp((port + 4), 11);  /* nDTR, nRTS, OUT1 for IRQs */
   outp(0x21, (inp(0x21) && 0xEF)); /* permit serial IRQs */
   outp((port + 1), 1);   /* Set IRQ on serial byte received */


   /* Open log file */
   fp = fopen("digibox.log", "wb");
   if (fp == NULL)
   {
      fprintf(stderr, "\"Digibox.log\" file couldn't be opened, this session \
will not be logged.\n");
   }
   else
   {
      timer = time(NULL);
      tblock = localtime(&timer);
      fprintf(fp, "; log started at %02d:%02d:%02d on %04d/%02d/%02d.\n",
              tblock->tm_hour, tblock->tm_min, tblock->tm_sec,
              (tblock->tm_year + 1900), (tblock->tm_mon + 1),
              tblock->tm_mday);
   }


   /* Now await data from the Digibox */
   cprintf("\r\n\r\nWaiting for data from the Digibox...\r\n\r\n");


   /* Set up the screen display */
   sleep(1);
   redraw_screen();
   cursor_off();


   /* Set up a catch on ^C keypresses */
   signal(SIGINT, ctrlc_catch);


   /* Main loop */
   do
   {
      /* We can outrun the Digibox, so incorporate a little delay. We use
         this icky code because "sleep()" is too long (a second) and when
         running within Windows, "delay()" (mS) can freeze the program... */
      for (key = 0; key < 8; key++)
      {
         tick = clock();          /* Wait for an 18.2th of a second (8x) */
         while (tick == clock()); /* That's about 2/5ths of a second...  */
      }


      /* Whizzy-bar "I'm waiting" indicator */
      gotoxy(1, 12);
      plotwhat[0] = activity[activone++];
      activone &= 3;
      puttext(80, 25, 80, 25, plotwhat);


      /* If there is data, deal with it! */
      if ( buffer_ready() != 0 )
      {
	 buffer_readstring();

	 write_raw_data();

	 if (fp != NULL)
         {
            /* This is optional code to push in a record of when
               the serial event occurred, in the logfile. Alter
               these comments if you wish to make use of it...
            timer = time(NULL);
            tblock = localtime(&timer);
            fprintf(fp, "; [%02d:%02d:%02d]\n",
              tblock->tm_hour, tblock->tm_min, tblock->tm_sec); */
	    fprintf(fp, "%s\n", command);
         }

	 decode_command();
      }


      /* Keypress? */
      if ( kbhit() )
      {
	 key = getch();

         switch (key)
         {
            case   0 : /* Control keypress */
                       key = getch();
                       switch (key)
                       {
                          case 25: /* Alt-P - print */
                                   dump_to_printer();
                                   break;

                          case 45: /* Alt-X - quit */
                                   write_misc_status(
                                     "Alt-X pressed, quitting... Bye bye!");
	                           exit(0);
                                   break; /* technically 'unreachable code' */

                          case 46: /* Alt-C - configure */
                                   /* TO BE ADDED */
                                   break;

                          case 59: /* F1 - help */
                                   display_help();
                                   break;
                       }
                       break;

            case  27 : /* ESCape - quit */
                       write_misc_status("Esc pressed, quitting... Bye bye!");
	               exit(0);
                       break; /* technically 'unreachable code' */

            case 't' : /* intentional fall-through to... */
            case 'T' : /* Serial transmit test [WON'T ACTUALLY WORK!!!] */
                       write_misc_status("Serial send test started...");
                       serial_senddata("M00");
                       write_misc_status("Serial send test finished.");
                       break;

            case '?' : /* intentional fall-through to... */
            case '/' : /* Redraw screen */
                       redraw_screen();
                       break;
         }
      }
   } while (TRUE);

#pragma warn -rch
   return 0; /* unreachable code, here to be tidy! */
#pragma warn +rch
}



void ctrlc_catch(int *reglist)
{
   /* Traps ^C presses; we need a controlled exit
      else the interrupt claim will be all messed up. */
   reglist = reglist;
   write_misc_status("Please use Esc or Alt-X to exit this software.");
   signal(SIGINT, ctrlc_catch);
   return;
}



void restore_vector(void)
{
   /* Restore serial port sanity, close file, normal screen. */

   /* Set serial port behaviour to normal */
   outp((port + 1), 0);   /* Disable IRQ on serial byte received */
   outp((port + 4), 0);   /* Disable DTR and RTS, no IRQs */

   /* Mask serial IRQs at 8259-alike */
   outp(0x21, (inp(0x21) || 0x10));

   /* Restore original serial handler */
   setvect(0x0C, oldserial);

   /* Close the log file */
   if (fp != NULL)
      fclose(fp);

   /* Restore the screen to normal */
   textbackground(BLACK);
   textcolor(LIGHTGRAY);
   gotoxy(1, 25);
   cprintf("\r\n");
   clreol();
   cprintf("\r\n");
   clreol();

   /* Print a bye-bye message [normal "printf", not via conio] */
   printf("\
Thank you for using this software. Don't forget you can check on my website to\n\
ensure that you are using the latest version of this software, as well as a\n\
host of other Digibox-related things . . . http://www.heyrick.co.uk/digibox/\
\n\n");

   return;
}



void interrupt serial_event()
{
   /* Get serial byte and write it to ring buffer.
      This is an interrupt routine ... MUST BE QUICK!
   */

   buffer[bufferhead++] = inp(port); /* no temp vars, read direct to buffer */
   if (bufferhead > 1023)
      bufferhead = 0;

   outp(0x20, 0x20);    /* Flag IRQ handled */

   return;
}



void serial_senddata(char *data)
{
   /* TEST - attempts to send a code to the Digibox...

      LATER: From a person who is distantly connected to the Pace
             digibox people, this stuff is purely an anachronism
             as the EPG system will NOT respond to these codes.
             [in other words, you might as well delete this]


      According to Libby's config files:
         Packet format:

         Byte:     1    2     3       4      5       6...
         Content  SYNC,SYNC,COMMAND,LABEL,BYTECOUNT,DATA...

      where:
         SYNC is 0xB1
         COMMAND is product specific command code
         LABEL is currently always zero
         BYTECOUNT is num of bytes in data field
         DATA is a variable num of bytes depending on packet
      so smallest packet is 5 bytes (bytecount=0)

      The potential problem is the COMMAND value. We'll try zero...
   */

   int  loop = 0;

   serial_writebyte(0xB1);
   serial_writebyte(0xB1);
   serial_writebyte(0);
   serial_writebyte(0);
   serial_writebyte(3);
   serial_writebyte('0');
   serial_writebyte('0');
   serial_writebyte('0');
   sleep(1);

   serial_writebyte(0xB1);
   serial_writebyte(0xB1);
   serial_writebyte(0);
   serial_writebyte(0);
   serial_writebyte(strlen(data));
   for (loop = 0; loop < strlen(data); loop++)
      serial_writebyte(data[loop]);
   return;
}



void serial_writebyte(int mybyte)
{
   /* Writes a byte on the serial port... */
   int  loop = 0;

   /* Wait for port to be ready */
   while ( ( (inp(port + 5) && 0x20) == 0 ) && ( loop < 1024 ) )
      loop++;

   /* Sanity check, so we don't lock up on a disconnected port... */
   if (loop > 1023)
   {
      write_misc_status("Serial transmit failed - tried 1024 times");
      return;
   }

   /* Send it */
   outp(port, mybyte);

   return;
}



int  buffer_ready(void)
{
   /* Returns whether or not there is anything in the input buffer.
      0 = nothing new; non-0 = data
   */

   return (bufferhead - buffertail);
}



int  buffer_readstring(void)
{
   /* Reads from the buffer into the 'command' string. The string is clipped
      when we reach an ASCII 10, or buffer head=tail. Returns bytes read.
   */

   int  posn = 0;
   int  done = FALSE;

   /* Trap no data */
   if (bufferhead == buffertail)
   {
      strcpy(command, "");
      return 0;
   }

   /* Else get a string */
   while ( done == FALSE )
   {
      /* Copy the byte */
      command[posn++] = buffer[buffertail++];
      if (buffertail > 1023)
	 buffertail = 0;

      /* If ASCII 10, fiddle for terminator */
      if (command[(posn - 1)] == 10)
      {
	 done = TRUE;
	 posn--;
      }

      /* End of data? */
      if (bufferhead == buffertail)
	 done = TRUE;

      /* Out of buffer space? */
      if (posn > 510)
	 done = TRUE;
   }

   /* Null terminate it for C... */
   command[posn] = '\0';

   /* Return bytes read */
   return posn;
}



void decode_command(void)
{
   /* Decodes the command string.

      There are two types of command string. The first starts with a '#'
      and marks when the interactive stuff begins and ends. The second
      starts with a three-digit number giving the length of the string.
   */

   char blockid[8] = "";
   char blockdata[512] = "";
   int  bposn = 0;
   int  cposn = 0;
   int  cstrt = 0;
   int  dposn = 0;
   int  loop = 0;

   if (command[0] == '#')
   {
      char seq[80] = "";

      sprintf(seq, "Main software event: %s.", command);
      write_misc_status(seq);
      return;
   }

   /* Command strings appear to be > 7 characters. Less we call invalid. */
   if (strlen(command) < 8)
      return;

   /* Check the command string length matches */
   if (strlen(command) != atoi(command))
      return;


   /* Command strings are in the form "SSCN010118". This means we are
      on TV channel 118. There can be numerous command strings chained
      together. The sequence ends with two characters, must likely some
      sort of checksum.
   */

   /* Knock off those last two characters */
   command[(strlen(command) - 1)] = '\0';

   /* Now step through it */
   cposn = 3; /* skip over command string length */
   while (cposn < strlen(command))
   {
      /* Get the block ID - this is a three/four digit character code */
      for (bposn = 0; bposn < 8; bposn++)
	 blockid[bposn] = '\0';
      bposn = 0;
      while ( isalpha(command[cposn]) )
      {
	 blockid[bposn++] = command[cposn++];
	 if (bposn > 7)
	 {
            write_misc_status(
              "Internal error: Block ID too long, skipping parse.");
	    return;
	 }
      }

      cstrt = cposn;

      /* Now look for 'SS'. */
      bposn = 0;
      while (bposn == 0)
      {
	 while ( (command[cposn] != 'S') && (cposn < strlen(command)) )
	    cposn++;

	 if (cposn >= strlen(command))
	    bposn = 1;

	 if (bposn == 0)
	 {
	    if (command[cposn + 1] == 'S')
	       bposn = 1;
	    cposn++;
	 }
      }

      /* Copy out string data */
      dposn = 0;
      for (loop = cstrt; loop <= (cposn - 2); loop++)
	 blockdata[dposn++] = command[loop];
      blockdata[dposn] = '\0';

      if (cposn != strlen(command))
	 cposn--; /* So can see next 'SS' marker */


      /* SSCN "###xxx" where "xxx" is channel number. No diff. for radio chnl.
	 i.e. "101984"

	 SSCA "###*" where "*" is channel name.
	 i.e. "016BBC R4 FM"

	 SSDT "### h.mmxx day no mon " where "h"our "mm"inute "xx"am/pm
		"day"name "no" of day "mon"th name"
	 i.e. "026 5.19pm Tue 27 Jun "

	 SST  "####h.mmxx" "h"our, "mm"inute "xx"am/pm of start of programme
	 i.e. "00135.00pm"  <-- But what happens for an 11pm programme?

	 SSN  "####x*y" where "*" is programme name. 'x' and 'y' are probably
	      quotes in some weird character set. ASCII 134, and 135.
	 i.e. "0011PM"

	 SSE  "####*x" where "*" is the programme description, and 'x' is a
	      byte that should be clipped off!
	 i.e. "0162Full coverage and analysis [...clip...] Then Weather.c"

         SSEI "###Autoview" - marks programme as being an autoview programme.


	 THE FOLLOWING ARE NOT PREFIXED WITH "SS" SO WE SEE THE ENTIRE STRING.

	 CE   "00010xxxy" is numerical keypress, and "xxx" is channel number.
	      The 'y' is a letter which may or may not be present.
	 i.e. "000109--a" -> "0001098-a" -> "00010984"
	      "00011xyyyz" where 'x' is ASCII 127, "yyy" is channel number.
	      The 'z' is the maybe/maybe-not letter. This is for RADIO chnls.
	 i.e. "00011x1--2" -> "00011x10-2" -> "00011x104".

         CEER "###Channel Unavailable" - this channel not in the EPG.

	 PUCP "###Audio Unavailable Please check your digital satellite
	       receiverx" where 'x' is an extra letter.
	 PUSP as for PUCP; these erroneous messages are given when the box is
	      tuned to a channel that either needs a card, or needs different
	      subscription... Also occurs if no satellite signal is coming in.

         PUPP "###An autoview event is about to start, to prevent channel
               change press Sky key"

	 SYIA "00809" seemed to be when returning from text

	 SYIC "xxx" where "xxx" is some sort of key code.
	      "001180803" - 'help' or 'yellow' pressed
	      "009--$"    - 'back up' pressed ($ is another letter) may repeat
	      "009--$"    - evidently also when 'text' pressed

	 SYFS "0081SYD1034Press Sky For Channel Audio$", occurs when the user
	      presses "tv guide", "box office", "services", or "interactive".
	      "0080$", occurs when "Sky" pressed to return to channel.

	 SYST "0081$" - Digibox put into standby
	      "0080$" - Digibox taken out of standby

      */

      bposn = 0;

      if ( strcmp(blockid, "SSCN") == 0 )
      {
	 write_channel_info((blockdata + 3), 1);
         bposn = 1;
      }

      if ( strcmp(blockid, "SSCA") == 0 )
      {
	 write_channel_info((blockdata + 3), 2);
         /* Only print 'received' if NOT alongside programme info */
         if (strstr(command, "SSN") == 0)
            write_misc_status("Received channel information.");
         bposn = 1;
      }

      if ( strcmp(blockid, "SSDT") == 0 )
      {
	 write_channel_info((blockdata + 3), 3);
         bposn = 1;
      }

      if ( strcmp(blockid, "SSN") == 0 )
      {
	 write_programme_info((blockdata + 4), 1);
         write_misc_status("Received programme information.");
         write_programme_info("", 3);
         bposn = 1;
      }

      if ( strcmp(blockid, "SST") == 0 )
      {
	 write_programme_info((blockdata + 4), 2);
         bposn = 1;
      }

      if ( strcmp(blockid, "SSEI") == 0 )
      {
	 write_programme_info("", 4);
         bposn = 1;
      }

      if ( (strcmp(blockid, "SSE") == 0) && (bposn == 0) )
      {
	 write_programme_info((blockdata + 4), 3);
         bposn = 1;
      }

      /* Anything else can be an 'event' */
      if (bposn == 0)
      {
         char seq[80] = "";

         if ( (strcmp(blockid, "CE") == 0) && (bposn == 0) )
         {
            if ( strstr(blockdata, "CEER") != 0 )
            {
               write_misc_status("Selected channel is not listed in EPG, \
reverting to previous.");
            }
            else
            {
               if (blockdata[4] == '0')
               {
                  sprintf(seq, "Remote pressed, TV channel selection: %s",
                          (blockdata + 5));
                  write_misc_status(seq);
               }
               else
               {
                  sprintf(seq, "Remote pressed, RADIO channel selection: %s",
                          (blockdata + 6));
                  write_misc_status(seq);
               }
            }
            bposn = 1;
         }

         if ( (strcmp(blockid, "PUCP")==0) || (strcmp(blockid, "PUSP")==0) )
         {
            write_misc_status("Channel unavailable (no suitable card, or no \
signal)");
            bposn = 1;
         }

         if ( strcmp(blockid, "PUPP") == 0 )
         {
            write_misc_status("About to switch to Autoview programme.");
            bposn = 1;
         }

         if ( strcmp(blockid, "SYIC") == 0 )
         {
            write_misc_status("'back up', 'help', 'yellow' or 'text' pressed.");
            bposn = 1;
         }

         if ( strcmp(blockid, "SYIA") == 0 )
         {
            write_misc_status("Returning to normal service from teletext or \
interactive.");
            bposn = 1;
         }

         if ( strcmp(blockid, "SYFS") == 0 )
         {
            write_misc_status("Function button ('tv guide', 'services', \
etc) pressed.");
            bposn = 1;
         }

         if ( strcmp(blockid, "SYST") == 0 )
         {
            if (blockdata[3] == '1')
               write_misc_status("Digital receiver switched into standby.");
            else
               write_misc_status("Digital receiver switched out of standby.");
            bposn = 1;
         }
      }

      /* Still not handled? */
      if (bposn == 0)
      {
         char seq[80] = "";

         sprintf(seq, "Command \"%s\" not recognised. Please inform me so \
I can update!", blockid);
         write_misc_status(seq);
      }
   }

   return;
}



void cursor_off(void)
{
   /* Turn off the cursor (generic video int method) */

   union REGS inreg, outreg;
   inreg.h.ah = 1;
   inreg.x.cx = 0x0F00;
   int86(0x10, &inreg, &outreg);
}



void cursor_on(void)
{
   /* Turn on the cursor again */

   union REGS inreg, outreg;
   inreg.h.ah = 1;
   inreg.x.cx = 0x0607;
   int86(0x10, &inreg, &outreg);
}



void redraw_screen(void)
{
   /* Draw the screen layout */

   int blocky = 0;
   int loop = 0;
   int recurse = 0;

   textbackground(BLACK);
   textcolor(WHITE);
   clrscr();

   /* Header */
   textbackground(WHITE);
   textcolor(YELLOW);
   cputs("Pace Serial Message Decoder       ");
   textcolor(LIGHTCYAN);
   cprintf("v%s        ", appvers);
   textcolor(LIGHTBLUE);
   cputs("http://www.heyrick.co.uk/digibox/");

   /* Draw the three boxes */
   for (blocky = 0; blocky < 3; blocky++)
   {
      /* Top row of box */
      textbackground(BLUE);
      textcolor(CYAN);
      putch(201);
      for (loop = 1; loop < 79; loop++)
	 putch(205);
      putch(187);
      /* Middle of box */
      for (recurse = 0; recurse < 5; recurse++)
      {
	 putch(186);
	 for (loop = 1; loop < 79; loop++)
	    putch(32);
	 putch(186);
      }
      /* Bottom row of 'raw data' box */
      putch(200);
      for (loop = 1; loop < 79; loop++)
	 putch(205);
      putch(188);
      if (blocky < 2)
	 cprintf("\r\n");
   }

   /* Footer */
   textbackground(WHITE);
   textcolor(BLACK);
   cprintf("Logfile is %s", ((fp==NULL) ? "inactive" : "active  "));
   textcolor(LIGHTBLUE);
   cputs("    Press [Esc] to exit, [F1] for help");
   clreol();
   /* Back up and add highlight */
   textcolor(WHITE);
   gotoxy(31, 25);
   cputs("Esc");
   gotoxy(46, 25);
   cputs("F1");

   /* Fill in the captions */
   textbackground(BLUE);
   textcolor(CYAN);
   gotoxy(4, 2);
   putch(185);
   textcolor(LIGHTCYAN);
   cputs(" Raw incoming data (--:--:--) ");
   textcolor(CYAN);
   putch(204);
   gotoxy(4, 10);
   putch(185);
   textcolor(LIGHTCYAN);
   cputs(" Decoded channel / programme information ");
   textcolor(CYAN);
   putch(204);
   gotoxy(4, 18);
   putch(185);
   textcolor(LIGHTCYAN);
   cputs(" Other status / activities ");
   textcolor(CYAN);
   putch(204);

   /* Channel/programme sections... */
   textcolor(LIGHTGREEN);
   gotoxy(3, 11);
   cputs("Channel #:       Name:                              Date: ");
   gotoxy(3, 12);
   cputs("On now:                                             Started at:");
   gotoxy(3, 13);
   cputs("Description:");

   return;
}



void write_raw_data(void)
{
   time_t timer;
   struct tm *tblock;
   int  loop = 0;

   timer = time(NULL);
   tblock = localtime(&timer);

   textbackground(BLUE);
   textcolor(LIGHTCYAN);
   gotoxy(24, 2);
   cprintf("(%02d:%02d:%02d)", tblock->tm_hour, tblock->tm_min,
                               tblock->tm_sec);

   window(3, 3, 78, 7);
   textcolor(WHITE);
   clrscr();
   for (loop = 0; loop < strlen(command); loop++)
   {
      if ( (command[loop] == 127) || (command[loop] < 32) )
      {
         cprintf("[%02X]", command[loop]);
      }
      else
      {
         putch(command[loop]);
      }
   }
   window(1, 1, 80, 25);
   return;
}



void write_channel_info(char *info, int type)
{
   /* Writes '1' channel number, '2' channel name, or '3' date. */

   textbackground(BLUE);
   textcolor(YELLOW);
   switch (type)
   {
      case 1: gotoxy(14, 11);
	      cputs("     ");
	      gotoxy(14, 11);
	      cputs(info);
	      break;

      case 2: gotoxy(26, 11);
	      cputs("                            ");
	      gotoxy(26, 11);
	      cputs(info);
	      break;

      case 3: info[strlen(info) - 1] = '\0';
	      gotoxy(61, 11);
	      cputs("                   ");
	      gotoxy(61, 11);
	      cputs(info);
	      break;
   }

   return;
}



void write_programme_info(char *info, int type)
{
   /* Writes '1' programme title, '2' start time, '3' description,
      or 4 if Autoview. */

   int  loop = 0;
   int  mybyte = 0;

   textbackground(BLUE);
   textcolor(YELLOW);
   switch (type)
   {
      case 1: gotoxy(11, 12);
	      cputs("                                            ");
	      gotoxy(11, 12);
              for (loop = 0; loop < strlen(info); loop++)
              {
                 mybyte = info[loop];
                 if (mybyte == 134) mybyte = '`';
                 if (mybyte == 135) mybyte = '\'';
                 putch(mybyte);
              }
	      break;

      case 2: gotoxy(67, 12);
	      cputs("        ");
	      gotoxy(67, 12);
	      cputs(info);
	      break;

      case 3: window(3, 13, 78, 15);
              clrscr();
              textcolor(LIGHTGREEN);
              cputs("Description: ");
              textcolor(YELLOW);
              write_programme_description(info);
              window(1, 1, 80, 25);
	      break;

      case 4: gotoxy(46, 11);
              textcolor(WHITE);
              cputs("AUTOVIEW");
              break;
   }

   return;
}



void write_programme_description(char *info)
{
   /* Including rudimentary word wrapping... */

   int  readposn = 0;
   int  lookahead = 0;
   int  infolen = 0;
   int  towrite = 0;

   infolen = strlen(info);

   while (readposn < infolen)
   {
      /* We need to read from "readposn" to the next SPACE (or end of string).
         Mark this position as "lookahead". Do the maths to work out how much
         needs to be written. If this will fit on the current line, output it
         and then make "readposn" equal "lookahead". If it will not fit,
         force a linefeed and then output it...
      */

      lookahead++; /* start with next character */
      while ( (info[lookahead] != ' ') && (lookahead < infolen) )
         lookahead++;

      towrite = lookahead - readposn;

      if ((wherex() + towrite) < 77)
      {
         /* it'll fit on this line */
         while (readposn < lookahead)
            putch(info[readposn++]);
      }
      else
      {
         /* Force a new line and start then... */
         cprintf("\r\n");
         while (info[readposn] == ' ')
            readposn++;
         while (readposn < lookahead)
            putch(info[readposn++]);
      }
   }

   return;
}


void write_misc_status(char *info)
{
   time_t timer;
   struct tm *tblock;

   timer = time(NULL);
   tblock = localtime(&timer);

   window(3, 19, 78, 23);
   if (ti.curx != 0)
   {
      /* Not the first call */
      gotoxy(ti.curx, ti.cury);
      cprintf("\r\n");
   }
   textbackground(BLUE);
   textcolor(WHITE);
   cprintf("%02d:%02d:%02d : %s", tblock->tm_hour, tblock->tm_min,
                                  tblock->tm_sec, info);
   gettextinfo(&ti);
   window(1, 1, 80, 25);
   return;
}



void display_help(void)
{
   textbackground(WHITE);
   clrscr();

   textcolor(YELLOW);
   textbackground(BLUE);
   clreol();
   cprintf("Pace Serial Message Decoder - v%s - %s - by %s           HELP",
           appvers, appdate, "Rick Murray");
   gotoxy(1, 25);
   clreol();
   cputs("                      Please press [SPACE] to continue...");

   /* First, write it normal */
   textbackground(WHITE);
   textcolor(BLACK);
   gotoxy(1, 3);
   cprintf("\
 This is PSMD - a serial status message decoder for use with digital\r\n\
 satellite receivers manufactured by Pace (i.e. BSkyB 2500B Digibox).\r\n\
\r\n\
 In order to use this software, connect a straight-through wired serial lead\r\n\
 (i.e. a standard 9-pin to 9-pin MODEM lead) between your computer and the\r\n\
 satellite receiver...\r\n\
\r\n\
\r\n\
\r\n\
\r\n\
\r\n\
 It is possible that, initially, nothing will appear on the screen. In this\r\n\
 case you should cause the Digibox to respond to you (I tend to press '1' and\r\n\
 then 'back up'). After that, it will notice the serial port is accepting data,\r\n\
 and will then start sending the status messages...\r\n\
\r\n\
 For further information on various aspects of Digibox operation and use, as\r\n\
 well as upgrades to this software, please visit:\r\n");
   /* The following is colourised */
   textcolor(BLUE);
   cputs("                       http://www.heyrick.co.uk/digibox/");
   textcolor(RED);
   cprintf("\r\n\r\n\
     IMPORTANT: THIS SOFTWARE IS NOT TO BE (RE)DISTRIBUTED FOR ANY REASON.\r\n\
                (refer to licence in the help file \"psmd.txt\" for details)");

   /* Now go back and fill in highlighted words */
   textcolor(BLUE);
   gotoxy(10, 3);
   cputs("PSMD");
   gotoxy(49, 4);
   cputs("BSkyB 2500B Digibox");
   gotoxy(34, 7);
   cputs("MODEM");
   gotoxy(72, 15);
   cputs("1");
   gotoxy(8, 16);
   cputs("back up");
   textbackground(BLUE);
   textcolor(WHITE);
   gotoxy(37, 25);
   cputs("Space");

   /* Drawing */
   textbackground(BLACK);
   textcolor(YELLOW);
   gotoxy(12, 9);
   cputs(" ͻ                               ͻ ");
   gotoxy(12, 10);
   cputs("                                                      ");
   gotoxy(12, 11);
   cputs("                                                      ");
   gotoxy(12, 12);
   cputs(" ͼ                               ͼ ");
   textcolor(GREEN + BLINK);
   gotoxy(25, 10);
   cputs("Ŀ  Ŀ  Ŀ  Ŀ  Ŀ   ");
   gotoxy(25, 11);
   cputs("           ");
   textcolor(WHITE);
   gotoxy(15, 10);
   cputs("Your");
   gotoxy(15, 11);
   cputs("computer");
   gotoxy(58, 10);
   cputs("Satellite");
   gotoxy(58, 11);
   cputs("receiver");

   getch();
   redraw_screen();

   return;
}



void get_true_version(void)
{
   /* Return TRUE version of DOS, pushes value into command buffer.
      [because Windows with SETVER can lie blatantly! :-) ] */
   int  maj = 0;
   int  min = 0;
   int  rev = 0;
   union REGS inreg, outreg;

   inreg.x.ax = 0x3306; /* actually clashes with some obscure thing! */
   int86(0x21, &inreg, &outreg);

   maj = outreg.h.bl;
   min = outreg.h.bh;
   rev = outreg.h.dl;

   sprintf(command, "%d.%02dr%d", maj, min, rev);

   return;
}



void dump_to_printer(void)
{
   char screen[ ((25 * 80) * 2) ];
   char printer[ (25 * 80) ];
   int  mybyte = 0;
   int  screenloop = 0;
   int  printerloop = 0;
   FILE *prnout = NULL;


   /* Get the information off of the screen */
   gettext(1, 1, 80, 25, screen);

   /* Copy the data from the screen data block to the printer */
   for (screenloop = 0; screenloop < ((25 * 80) * 2); screenloop += 2)
   {
      mybyte = screen[screenloop];

      /* Strip control codes */
      if ( (mybyte < 32) || (mybyte == 127) )
         mybyte = '.';

      /* Translate DOS box characters to plain ASCII */
      switch (mybyte)
      {
         case 187 : /* Top left and top right */
         case 201 : mybyte = '.';
                    break;

         case 188 : /* Bottom left and bottom right */
         case 200 : mybyte = '\'';
                    break;

         case 185 : /* Left and right T-piece */
         case 204 : mybyte = ':';
                    break;

         case 186 : /* Vertical */
                    mybyte = '|';
                    break;

         case 205 : /* Horizontal */
                    mybyte = '-';
                    break;
      }

      printer[printerloop++] = mybyte;
   }

   /* Now write it to a file on disc */
   prnout = fopen("prnout.txt", "wb");
   if (prnout == NULL)
   {
      write_misc_status("Can't create print file, printing aborted.");
      return;
   }

   for (printerloop = 0; printerloop < 25; printerloop++)
   {
      for (screenloop = 0; screenloop < 80; screenloop++)
      {
         fputc(printer[ ((printerloop * 80) + screenloop) ], prnout);
      }
      fprintf(prnout, "\r\n");
   }
   fputc(12, prnout); /* Force a form feed */
   fclose(prnout);

   textcolor(LIGHTCYAN);
   textbackground(BLACK);
   clrscr();
   cprintf("Now sending print file (\"prnout.txt\") to LPT1.\r\n\r\n");
   cprintf("\
If you see an error (possible, if your printer connects via USB or the\r\n\
like), please choose to ABORT the print job. You will then need to load\r\n\
the file into Notepad, and print it through Windows. It depends upon your\r\n\
printer configuration and driver capabilities.\r\n\r\n");
   textcolor(WHITE);

   /* This was necessary because it is no longer possible
      to simply interrogate parallel flags; doubly so if
      something picks up and queues LPT: output.
      It would have been nice if my Lexmark Z23 picked
      up on PRN: but trying to use PRN: crashed (!). I do
      not know if the driver is buggy, or Windows. It was
      Spool32 that died (no surprise there, then!). */
   printerloop = system("copy prnout.txt lpt1 /a");

   /* We conveniently have a copy of the screen we can push back! */
   clrscr();
   puttext(1, 1, 80, 25, screen);

   write_misc_status("Copy of screen sent to printer");

   return;
}
